#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "xmlrpc_format.h"

void test1()
{
    char *expect = "<methodCall><methodName>test1</methodName><params><param><value><i4>123</i4></value></param><param><value><i4>456</i4></value></param></params></methodCall>";
    char actual[1000];
    xmlrpc_request(actual, sizeof(actual), "test1", 2, xmlrpc_i4(123), xmlrpc_i4(456));
    if(strcmp(actual, expect) != 0)
    {
        printf("Format i4 failure\n");
        printf("  expect : %s\n", expect);
        printf("  actual : %s\n", actual);
    }
}

void test2()
{
    char *expect = "<methodCall><methodName>test2</methodName><params><param><value><string>ABC</string></value></param><param><value><string>DEF</string></value></param></params></methodCall>";
    char actual[1000];
    xmlrpc_request(actual, sizeof(actual), "test2", 2, xmlrpc_string("ABC"), xmlrpc_string("DEF"));
    if(strcmp(actual, expect) != 0)
    {
        printf("Format string failure\n");
        printf("  expect : %s\n", expect);
        printf("  actual : %s\n", actual);
    }
}

void test3()
{
    char *expect = "<methodCall><methodName>test3</methodName><params><param><value><struct><member><name>ival</name><value><i4>123</i4></value></member><member><name>sval</name><value><string>ABC</string></value></member></struct></value></param></params></methodCall>";
    char actual[1000];
    xmlrpc_request(actual, sizeof(actual), "test3", 1, xmlrpc_struct(2, "ival", xmlrpc_i4(123), "sval", xmlrpc_string("ABC")));
    if(strcmp(actual, expect) != 0)
    {
        printf("Format struct failure\n");
        printf("  expect : %s\n", expect);
        printf("  actual : %s\n", actual);
    }
}

void test4()
{
    char *expect = "<methodCall><methodName>test4</methodName><params><param><value><array><data><value><i4>123</i4></value><value><i4>456</i4></value></data></array></value></param></params></methodCall>";
    char actual[1000];
    xmlrpc_request(actual, sizeof(actual), "test4", 1, xmlrpc_array(2, xmlrpc_i4(123), xmlrpc_i4(456)));
    if(strcmp(actual, expect) != 0)
    {
        printf("Format array i4 failure\n");
        printf("  expect : %s\n", expect);
        printf("  actual : %s\n", actual);
    }
}

void test5()
{
    char *expect = "<methodCall><methodName>test5</methodName><params><param><value><array><data><value><string>ABC</string></value><value><string>DEF</string></value></data></array></value></param></params></methodCall>";
    char actual[1000];
    xmlrpc_request(actual, sizeof(actual), "test5", 1, xmlrpc_array(2, xmlrpc_string("ABC"), xmlrpc_string("DEF")));
    if(strcmp(actual, expect) != 0)
    {
        printf("Format array string failure\n");
        printf("  expect : %s\n", expect);
        printf("  actual : %s\n", actual);
    }
}

int main()
{
    test1();
    test2();
    test3();
    test4();
    test5();
    return 0;
}

