#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "xmlrpc_parse.h"

void test1(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        int v;
        xmlrpc_response(xmlstr, &v);
        if(v != 123)
        {
            printf("Parse i4 failure\n");
            printf("  expect : 123\n");
            printf("  actual : %d\n", v);
        }
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        printf("  error: %d %s\n", errcod, errmsg);
        free(errmsg);
    }
}

void test2(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        char *s;
        xmlrpc_response(xmlstr, &s);
        if(strcmp(s, "ABC") != 0)
        {
            printf("Parse string failure\n");
            printf("  expect : ABC\n");
            printf("  actual : %s\n", s);
        }
        free(s);
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        printf("  error: %d %s\n", errcod, errmsg);
        free(errmsg);
    }
}

void test3(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        int *ia = malloc(1000 * sizeof(int));
        int n = xmlrpc_response(xmlstr, ia);
        if(n != 2)
        {
            printf("Parse array i4 num element failure\n");
            printf("  expect : 2\n");
            printf("  actual : %d\n", n);
        }
        if(ia[0] != 123)
        {
            printf("Parse array i4 element 1 failure\n");
            printf("  expect : 123\n");
            printf("  actual : %d\n", ia[0]);
        }
        if(ia[1] != 456)
        {
            printf("Parse array i4 element 2 failure\n");
            printf("  expect : 456\n");
            printf("  actual : %d\n", ia[1]);
        }
        free(ia);
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        printf("  error: %d %s\n", errcod, errmsg);
        free(errmsg);
    }
}

void test4(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        char **sa = malloc(1000 * sizeof(char *));
        int n = xmlrpc_response(xmlstr, sa);
        if(n != 2)
        {
            printf("Parse array string num element failure\n");
            printf("  expect : 2\n");
            printf("  actual : %d\n", n);
        }
        if(strcmp(sa[0], "ABC") != 0)
        {
            printf("Parse array string element 1 failure\n");
            printf("  expect : ABC\n");
            printf("  actual : %s\n", sa[0]);
        }
        if(strcmp(sa[1], "DEF") != 0)
        {
            printf("Parse array string element 2 failure\n");
            printf("  expect : DEF\n");
            printf("  actual : %s\n", sa[1]);
        }
        free(sa);
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        printf("  error: %d %s\n", errcod, errmsg);
        free(errmsg);
    }
}

struct data
{
    int ival;
    char *sval;
};

void test5(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        struct xmlrpc_struct_i4_or_str gas[2];
        xmlrpc_response(xmlstr, gas);
        struct data d;
        for(int i = 0; i < 2; i++)
        {
            if(strcmp(gas[i].name, "ival") == 0)
            {
                d.ival = gas[i].value.i4;
            }
            if(strcmp(gas[i].name, "sval") == 0)
            {
                d.sval = gas[i].value.str;
            }
        }
        if(d.ival != 123)
        {
            printf("Parse struct ival failure\n");
            printf("  expect : 123\n");
            printf("  actual : %d\n", d.ival);
        }
        if(strcmp(d.sval, "ABC") != 0)
        {
            printf("Parse struct sval failure\n");
            printf("  expect : ABC\n");
            printf("  actual : %s\n", d.sval);
        }
        free(d.sval);
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        printf("  error: %d %s\n", errcod, errmsg);
        free(errmsg);
    }
}

void test6(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        struct xmlrpc_struct_i4_or_str *gas = malloc(2000 * sizeof(struct xmlrpc_struct_i4_or_str));
        int n = xmlrpc_response(xmlstr, gas);
        struct data *da = malloc(1000 * sizeof(struct data));
        for(int i = 0; i < n; i++)
        {
            if(strcmp(gas[i].name, "ival") == 0)
            {
                da[i / 2].ival = gas[i].value.i4;
            }
            if(strcmp(gas[i].name, "sval") == 0)
            {
                da[i / 2].sval = gas[i].value.str;
            }
        }
        free(gas);
        if(n != 4)
        {
            printf("Parse array struct num element failure\n");
            printf("  expect : 4\n");
            printf("  actual : %d\n", n);
        }
        if(da[0].ival != 123)
        {
            printf("Parse array struct element 1 ival failure\n");
            printf("  expect : 123\n");
            printf("  actual : %d\n", da[0].ival);
        }
        if(strcmp(da[0].sval, "ABC") != 0)
        {
            printf("Parse array struct element 1 sval failure\n");
            printf("  expect : ABC\n");
            printf("  actual : %s\n", da[0].sval);
        }
        free(da[0].sval);
        if(da[1].ival != 456)
        {
            printf("Parse array struct element 2 ival failure\n");
            printf("  expect : 456\n");
            printf("  actual : %d\n", da[1].ival);
        }
        if(strcmp(da[1].sval, "DEF") != 0)
        {
            printf("Parse array struct element 2 sval failure\n");
            printf("  expect : DEF\n");
            printf("  actual : %s\n", da[1].sval);
        }
        free(da[1].sval);
        free(da);
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        printf("  error: %d %s\n", errcod, errmsg);
        free(errmsg);
    }
}

void test7(const char *xmlstr)
{
    if(xmlrpc_response_status(xmlstr))
    {
        printf("Ooops\n");
    }
    else
    {
        int errcod;
        char *errmsg;
        xmlrpc_response_errmsg(xmlstr, &errcod, &errmsg);
        if(errcod != 101)
        {
            printf("Parse error code failure\n");
            printf("  expect : 101\n");
            printf("  actual : %d\n", errcod);
        }
        if(strcmp(errmsg, "Expected error") != 0)
        {
            printf("Parse error message failure\n");
            printf("  expect : Expected error\n");
            printf("  actual : %s\n", errmsg);
        }
        free(errmsg);
    }
}

int main()
{
    test1("<methodResponse>\n"
          "  <params>\n"
          "    <param>\n"
          "      <value><i4>123</i4></value>\n"
          "    </param>\n"
          "  </params>\n"
          "</methodResponse>");
    test2("<methodResponse>\n"
          "  <params>\n"
          "    <param>\n"
          "      <value><string>ABC</string></value>\n"
          "    </param>\n"
          "  </params>\n"
          "</methodResponse>");
    test3("<methodResponse>\n"
          "  <params>\n"
          "    <param>\n"
          "      <value>\n"
          "        <array>\n"
          "          <data>\n"
          "            <value><i4>123</i4></value>\n"
          "            <value><i4>456</i4></value>\n"
          "          </data>\n"
          "        </array>\n"
          "      </value>\n"
          "    </param>\n"
          "  </params>\n"
          "</methodResponse>");
    test4("<methodResponse>\n"
          "  <params>\n"
          "    <param>\n"
          "      <value>\n"
          "        <array>\n"
          "          <data>\n"
          "            <value><string>ABC</string></value>\n"
          "            <value><string>DEF</string></value>\n"
          "          </data>\n"
          "        </array>\n"
          "      </value>\n"
          "    </param>\n"
          "  </params>\n"
          "</methodResponse>");
    test5("<methodResponse>\n"
          "  <params>\n"
          "    <param>\n"
          "      <value>\n"
          "        <struct>\n"
          "          <member>\n"
          "            <name>ival</name>\n"
          "            <value><i4>123</i4></value>\n"
          "          </member>\n"
          "          <member>\n"
          "            <name>sval</name>\n"
          "            <value><string>ABC</string></value>\n"
          "          </member>\n"
          "        </struct>\n"
          "      </value>\n"
          "    </param>\n"
          "  </params>\n"
          "</methodResponse>");
    test6("<methodResponse>\n"
          "  <params>\n"
          "    <param>\n"
          "      <value>\n"
          "        <array>\n"
          "          <data>\n"
          "            <struct>\n"
          "              <member>\n"
          "                <name>ival</name>\n"
          "                <value><i4>123</i4></value>\n"
          "              </member>\n"
          "              <member>\n"
          "                <name>sval</name>\n"
          "                <value><string>ABC</string></value>\n"
          "              </member>\n"
          "            </struct>\n"
          "            <struct>\n"
          "              <member>\n"
          "                <name>ival</name>\n"
          "                <value><i4>456</i4></value>\n"
          "              </member>\n"
          "              <member>\n"
          "                <name>sval</name>\n"
          "                <value><string>DEF</string></value>\n"
          "              </member>\n"
          "            </struct>\n"
          "          </data>\n"
          "        </array>\n"
          "      </value>\n"
          "    </param>\n"
          "  </params>\n"
          "</methodResponse>");
    test7("<methodResponse>\n"
          "  <fault>\n"
          "    <value>\n"
          "      <struct>\n"
          "        <member>\n"
          "          <name>faultCode</name>\n"
          "          <value><int>101</int></value>\n"
          "        </member>\n"
          "        <member>\n"
          "          <name>faultString</name>\n"
          "          <value><string>Expected error</string></value>\n"
          "        </member>\n"
          "      </struct>\n"
          "    </value>\n"
          "  </fault>\n"
          "</methodResponse>");
    test1("<methodResponse>"
          "<params>"
          "<param>"
          "<value><i4>123</i4></value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test2("<methodResponse>"
          "<params>"
          "<param>"
          "<value><string>ABC</string></value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test3("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<array>"
          "<data>"
          "<value><i4>123</i4></value>"
          "<value><i4>456</i4></value>"
          "</data>"
          "</array>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test4("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<array>"
          "<data>"
          "<value><string>ABC</string></value>"
          "<value><string>DEF</string></value>"
          "</data>"
          "</array>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test5("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<struct>"
          "<member>"
          "<name>ival</name>"
          "<value><i4>123</i4></value>"
          "</member>"
          "<member>"
          "<name>sval</name>"
          "<value><string>ABC</string></value>"
          "</member>"
          "</struct>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test6("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<array>"
          "<data>"
          "<struct>"
          "<member>"
          "<name>ival</name>"
          "<value><i4>123</i4></value>"
          "</member>"
          "<member>"
          "<name>sval</name>"
          "<value><string>ABC</string></value>"
          "</member>"
          "</struct>"
          "<struct>"
          "<member>"
          "<name>ival</name>"
          "<value><i4>456</i4></value>"
          "</member>"
          "<member>"
          "<name>sval</name>"
          "<value><string>DEF</string></value>"
          "</member>"
          "</struct>"
          "</data>"
          "</array>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test7("<methodResponse>"
          "<fault>"
          "<value>"
          "<struct>"
          "<member>"
          "<name>faultCode</name>"
          "<value><int>101</int></value>"
          "</member>"
          "<member>"
          "<name>faultString</name>"
          "<value><string>Expected error</string></value>"
          "</member>"
          "</struct>"
          "</value>"
          "</fault>"
          "</methodResponse>");
    test2("<methodResponse>"
          "<params>"
          "<param>"
          "<value>ABC</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test4("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<array>"
          "<data>"
          "<value>ABC</value>"
          "<value>DEF</value>"
          "</data>"
          "</array>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test5("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<struct>"
          "<member>"
          "<name>ival</name>"
          "<value><i4>123</i4></value>"
          "</member>"
          "<member>"
          "<name>sval</name>"
          "<value>ABC</value>"
          "</member>"
          "</struct>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    test6("<methodResponse>"
          "<params>"
          "<param>"
          "<value>"
          "<array>"
          "<data>"
          "<struct>"
          "<member>"
          "<name>ival</name>"
          "<value><i4>123</i4></value>"
          "</member>"
          "<member>"
          "<name>sval</name>"
          "<value>ABC</value>"
          "</member>"
          "</struct>"
          "<struct>"
          "<member>"
          "<name>ival</name>"
          "<value><i4>456</i4></value>"
          "</member>"
          "<member>"
          "<name>sval</name>"
          "<value>DEF</value>"
          "</member>"
          "</struct>"
          "</data>"
          "</array>"
          "</value>"
          "</param>"
          "</params>"
          "</methodResponse>");
    return 0;
}
