#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <descrip.h>

#include "vms_socket.h"
#include "vms_http.h"

static char *in_cstr(struct dsc$descriptor *s)
{
    int slen, ix;
    char *sptr, *res;
    if(s->dsc$b_class == DSC$K_CLASS_S || s->dsc$b_class == DSC$K_CLASS_D)
    {
        slen = s->dsc$w_length;
        sptr = s->dsc$a_pointer;
    }
    else if(s->dsc$b_class == DSC$K_CLASS_VS)
    {
        slen = *((short int *)s->dsc$a_pointer);
        sptr = s->dsc$a_pointer + sizeof(short int);
    }
    else
    {
        fprintf(stderr, "Unsupported descriptor class: %d\n", s->dsc$b_class);
        exit(44);
    }
    res = malloc(slen + 1);
    strncpy(res, sptr, slen);
    res[slen] = 0;
    /* trim traling whitespace */
    ix = slen - 1;
    while(ix >= 0 && (res[ix] == ' ' || res[ix] == 0))
    {
        res[ix] = 0;
        ix--;
    }
    return res;
}

static void format_get_head(char *head,
                            const char *path,
                            const char *host,
                            const char *acctyp)
{
   sprintf(head, "GET %s HTTP/1.1\r\n"
                 "Host: %s\r\n"
                 "Accept: %s\r\n"
                 "Connection: close\r\n"
                 "User-Agent: VMS_HTTP\r\n"
                 "\r\n", path, host, acctyp);
}

static void format_post_head(char *head,
                             const char *path,
                             const char *host,
                             const char *contyp,
                             int bodylen,
                             const char *acctyp)
{
   sprintf(head, "POST %s HTTP/1.1\r\n"
                 "Host: %s\r\n"
                 "Content-Type: %s\r\n"
                 "Content-Length: %d\r\n"
                 "Accept: %s\r\n"
                 "Connection: close\r\n"
                 "User-Agent: VMS_HTTP\r\n"
                 "\r\n", path, host, contyp, bodylen, acctyp);
}

static const char *TEH_C = "transfer-encoding: chunked";
static const char *CLH = "content-length: ";
static const char *CTH = "content-type: ";

static void skip_head(int sd, int *chunked, int *length, int *numcode,
                      char *txtcode, int txtcodelen, char *contyp, int contyplen)
{
   char buf[1000];
   int i;
   socket_recv(sd, buf, sizeof(buf));
   sscanf(buf + 9, "%d", numcode);
   strncpy(txtcode, buf + 13, txtcodelen);
   *chunked = 0;
   do
   {
      socket_recv(sd, buf, sizeof(buf));
      for(i = 0; i < strlen(buf); i++)
      {
         buf[i] = tolower(buf[i]);
      }
      if(strstr(buf, TEH_C) == buf)
      {
          *chunked = 1;
      }
      if(strstr(buf, CLH) == buf)
      {
         sscanf(buf + strlen(CLH), "%d", length);
      }
      if(strstr(buf, CTH) == buf) 
      {
         strncpy(contyp, buf + strlen(CTH), contyplen);
      }
   } while(strlen(buf) > 0);
}

struct vms_http *http_get(const char *host, int port, const char *path,
                          const char *acctyp)
{
   struct vms_http *con;
   char host2[1000],head[10000];
   con = (struct vms_http *)malloc(sizeof(struct vms_http));
   memset(con, 0, sizeof(struct vms_http));
   con->sd = socket_open(host, port);
   sprintf(host2, "%s:%d", host, port);
   format_get_head(head, path, host2, acctyp);
   socket_send(con->sd, head);
   skip_head(con->sd, &con->chunked, &con->length, &con->numcode,
             con->txtcode, sizeof(con->txtcode), con->contyp, sizeof(con->contyp));
   return con;
}

struct vms_http *http_post(const char *host, int port, const char *path,
                           const char *contyp, const char *acctyp, const char *body)
{
   struct vms_http *con;
   char host2[1000],head[10000];
   con = (struct vms_http *)malloc(sizeof(struct vms_http));
   memset(con, 0, sizeof(struct vms_http));
   con->sd = socket_open(host, port);
   sprintf(host2, "%s:%d", host, port);
   format_post_head(head, path, host2, contyp, strlen(body), acctyp);
   socket_send(con->sd, head);
   socket_send(con->sd, body);
   skip_head(con->sd, &con->chunked, &con->length, &con->numcode,
             con->txtcode, sizeof(con->txtcode), con->contyp, sizeof(con->contyp));
   return con;
}

int http_recv(struct vms_http *con, char *inmsg, int inmsgsiz, short int *inmsglen)
{
   int term;
   char dummy[100];
   *inmsglen = socket_recv_ext(con->sd, inmsg, inmsgsiz, &term);
   if(*inmsglen < 0) return -1;
   if(con->chunked)
   {
      /* start chunk */
      if(con->length == 0)
      {
         sscanf(inmsg, "%x", &con->length);
         if(con->length == 0) return -1;
         con->count = 0;
         *inmsglen = socket_recv_ext(con->sd, inmsg, inmsgsiz, &term);
      }
      /* end chunk */
      if(con->count + (*inmsglen + term) >= con->length)
      {
         con->length = 0;
      }
   }
   con->count += (*inmsglen + term);
   return *inmsglen;
}

int http_recv_all(struct vms_http *con, char *inmsg, int inmsgsiz, short int *inmsglen)
{
    int ix;
    short int len;
    ix = 0;
    while(http_recv(con, inmsg + ix, inmsgsiz - ix, &len) > 0)
    {
        ix += len;
    }
    *inmsglen = ix;
    return ix;
}

void http_close(struct vms_http *con)
{
   socket_close(con->sd);
   free(con);
}

int http_length(struct vms_http *con)
{
   return con->length;
}

int http_numcode(struct vms_http *con)
{
   return con->numcode;
}

char *http_txtcode(struct vms_http *con)
{
   return con->txtcode;
}

char *http_contyp(struct vms_http *con)
{
   return con->contyp;
}

struct vms_http *vms_http_get(struct dsc$descriptor *host, int *port, struct dsc$descriptor *path,
                              struct dsc$descriptor *acctyp)
{
   struct vms_http *con;
   char *path2, *host2, *acctyp2;
   host2 = in_cstr(host);
   path2 = in_cstr(path);
   acctyp2 = in_cstr(acctyp);
   con = http_get(host2, *port, path2, acctyp2);
   free(host2);
   free(path2);
   free(acctyp2);
   return con;
}

struct vms_http *vms_http_post(struct dsc$descriptor *host, int *port, struct dsc$descriptor *path,
                               struct dsc$descriptor *contyp, struct dsc$descriptor *acctyp, struct dsc$descriptor *body)
{
   struct vms_http *con;
   char *path2, *host2, *contyp2, *acctyp2, *body2;
   host2 = in_cstr(host);
   path2 = in_cstr(path);
   contyp2 = in_cstr(contyp);
   acctyp2 = in_cstr(acctyp);
   body2 = in_cstr(body);
   con = http_post(host2, *port, path2, contyp2, acctyp2, body2);
   free(host2);
   free(path2);
   free(contyp2);
   free(acctyp2);
   free(body2);
   return con;
}

int vms_http_recv(struct vms_http **con, struct dsc$descriptor *inmsg, short int *inmsglen)
{
   return http_recv(*con, inmsg->dsc$a_pointer, inmsg->dsc$w_length, inmsglen);
}

int vms_http_recv_all(struct vms_http **con, struct dsc$descriptor *inmsg, short int *inmsglen)
{
   return http_recv_all(*con, inmsg->dsc$a_pointer, inmsg->dsc$w_length, inmsglen);
}

void vms_http_close(struct vms_http **con)
{
   http_close(*con);
}

int vms_http_length(struct vms_http **con)
{
   return http_length(*con);
}

int vms_http_numcode(struct vms_http **con)
{
   return http_numcode(*con);
}

void vms_http_txtcode(struct vms_http **con, struct dsc$descriptor *txtcode, short int *txtcodelen)
{
   *txtcodelen = strlen((*con)->txtcode);
   strncpy(txtcode->dsc$a_pointer, (*con)->txtcode, *txtcodelen);
}

void vms_http_contyp(struct vms_http **con, struct dsc$descriptor *contyp, short int *contyplen)
{
   *contyplen = strlen((*con)->contyp);
   strncpy(contyp->dsc$a_pointer, (*con)->contyp, *contyplen);
}

