#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/socket.h>
#include <netdb.h>
#include <unixio.h>

#include <descrip.h>

#include "vms_socket.h"

int socket_open(const char *host, int port)
{
    int sd, stat;
    char sport[11];
    struct addrinfo hints, *res;
    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;
    hints.ai_flags = 0;
    sprintf(sport, "%d", port);
    stat = getaddrinfo(host, sport, &hints, &res);
    if(stat != 0) return -1;
    sd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if(sd < 0) return sd;
    stat = connect(sd, res->ai_addr, res->ai_addrlen);
    if(stat != 0) return -1;
    return sd;
}

int socket_send(int sd, const char *outmsg)
{
    int stat;
    stat = send(sd, outmsg, strlen(outmsg), 0);
    return stat;
}

int socket_recv(int sd, char *inmsg, int inmsglen)
{
   int term;
   return socket_recv_ext(sd, inmsg, inmsglen, &term);
}

int socket_recv_all(int sd, char *inmsg, int inmsglen)
{
    int ix, len;
    ix = 0;
    while((len = socket_recv(sd, inmsg + ix, inmsglen - ix)) > 0)
    {
        ix += len;
    }
    return ix;
}

int socket_recv_ext(int sd, char *inmsg, int inmsglen, int *term)
{
    int ix, len;
    ix = 0;
    inmsg[ix] = '\0';
    while(ix < inmsglen && strstr(inmsg, "\n") == NULL)
    {
        len = recv(sd, inmsg + ix, 1, 0);
        if(len == 0)
        {
            if(strlen(inmsg) == 0)
            {
                return -1;
            }
            else
            {
                break;
            }
        }
        ix = ix + len;
        inmsg[ix] = 0;
    }
    *term = 0;
    while(ix > 0 && (inmsg[ix - 1] == '\n' || inmsg[ix - 1] == '\r'))
    {
        inmsg[ix - 1] = 0;
        ix--;
        *term = *term + 1;
    }
    return ix;
}

void socket_close(int sd)
{
    close(sd);
}

static char *in_cstr(struct dsc$descriptor *s)
{
    int slen, ix;
    char *sptr, *res;
    if(s->dsc$b_class == DSC$K_CLASS_S || s->dsc$b_class == DSC$K_CLASS_D)
    {
        slen = s->dsc$w_length;
        sptr = s->dsc$a_pointer;
    }
    else if(s->dsc$b_class == DSC$K_CLASS_VS) 
    {
        slen = *((short int *)s->dsc$a_pointer);
        sptr = s->dsc$a_pointer + sizeof(short int);
    }
    else
    {
        fprintf(stderr, "Unsupported descriptor class: %d\n", s->dsc$b_class);
        exit(44);
    }
    res = malloc(slen + 1);
    strncpy(res, sptr, slen);
    res[slen] = 0;
    /* trim traling whitespace */
    ix = slen - 1;
    while(ix >= 0 && (res[ix] == ' ' || res[ix] == 0))
    {
        res[ix] = 0;
        ix--;
    }
    return res;
}

int vms_socket_open(struct dsc$descriptor *host, int *port)
{
    int sd;
    char *host2;
    host2 = in_cstr(host);
    sd = socket_open(host2, *port);
    free(host2);
    return sd;
}

int vms_socket_send(int *sd, struct dsc$descriptor *outmsg)
{
    int stat;
    char *outmsg2;
    outmsg2 = in_cstr(outmsg);
    stat = socket_send(*sd, outmsg2);
    socket_send(*sd, "\r\n"); /* add CR LF */
    free(outmsg2);
    return stat;
}

int vms_socket_recv(int *sd, struct dsc$descriptor *inmsg, short int *inmsglen)
{
   int term;
   return vms_socket_recv_ext(sd, inmsg, inmsglen, &term);
}

int vms_socket_recv_ext(int *sd, struct dsc$descriptor *inmsg, short int *inmsglen, int *term)
{
    int stat;
    stat = socket_recv_ext(*sd, inmsg->dsc$a_pointer, inmsg->dsc$w_length, term);
    *inmsglen = strlen(inmsg->dsc$a_pointer);
    return stat;
}

void vms_socket_close(int *sd)
{
    socket_close(*sd);
}
