#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "xmlrpc_format.h"

static const int SMALL = 100;
static const int MEDIUM = 10000;
static const int LARGE = 100000;
static const int XLARGE = 1000000;

static const char *i4_fmt = "<i4>%d</i4>";

char *xmlrpc_i4(int val)
{
    char *res = malloc(SMALL);
    snprintf(res, SMALL, i4_fmt, val);
    return res;
}

static const char *string_fmt = "<string>%s</string>";

char *xmlrpc_string(const char *val)
{
    char *res = malloc(LARGE);
    snprintf(res, LARGE, string_fmt, val);
    return res;
}

static const char *member_fmt = "<member><name>%s</name><value>%s</value></member>";

char *xmlrpc_member(const char *nam, char *val)
{
    char *res = malloc(LARGE);
    snprintf(res, LARGE, member_fmt, nam, val);
    free(val);
    return res;
}

static const char *struct_fmt = "<struct>%s</struct>";

char *xmlrpc_struct(int nmemb, ...)
{
    char *res = malloc(XLARGE);
    char *temp = malloc(XLARGE);
    strcpy(temp, "");
    va_list ap;
    va_start(ap, nmemb);
    for(int i = 0; i < nmemb; i++) 
    {
        const char *nam = va_arg(ap, const char *);
        char *val = va_arg(ap, char *);
        char *p = xmlrpc_member(nam, val);
        strcat(temp, p);
        free(p);
    }
    snprintf(res, XLARGE, struct_fmt, temp);
    free(temp);
    return res;
}

char *xmlrpc_struct2(int nmemb, ...)
{
    char *res = malloc(XLARGE);
    char *temp = malloc(XLARGE);
    strcpy(temp, "");
    va_list ap;
    va_start(ap, nmemb);
    for(int i = 0; i < nmemb; i++)
    {
        char *memb = va_arg(ap, char *);
        strcat(temp, memb);
    }
    snprintf(res, XLARGE, struct_fmt, temp);
    free(temp);
    return res;
}

static const char *value_fmt = "<value>%s</value>";

char *xmlrpc_value(char *val)
{
    char *res = malloc(LARGE);
    snprintf(res, LARGE, value_fmt, val);
    free(val);
    return res;
}

static const char *array_fmt = "<array><data>%s</data></array>";

char *xmlrpc_array(int nelm, ...)
{
    char *res = malloc(XLARGE);
    char *temp = malloc(XLARGE);
    strcpy(temp, "");
    va_list ap;
    va_start(ap, nelm);
    for(int i = 0; i < nelm; i++) 
    {
        char *val = va_arg(ap, char *);
        char *p = xmlrpc_value(val);
        strcat(temp, p);
        free(p);
    }
    snprintf(res, XLARGE, array_fmt, temp);
    free(temp);
    return res;
}

static const char *param_fmt = "<param>%s</param>";

char *xmlrpc_param(char *p)
{
    char *res = malloc(XLARGE);
    char *vp = xmlrpc_value(p);
    snprintf(res, XLARGE, param_fmt, vp);
    free(vp);
    free(p);
    return res;
}

static const char *request_fmt = "<methodCall><methodName>%s</methodName><params>%s</params></methodCall>";

static void xmlrpc_request_real(char *buf, int bufsiz, const char *func, int narg, va_list ap)
{
    char *temp = malloc(bufsiz);
    strcpy(temp, "");
    for(int i = 0;i < narg; i++) 
    {
        char *p = va_arg(ap, char *);
        char *p2 = xmlrpc_param(p);
        strcat(temp, p2);
        free(p2);
    }
    snprintf(buf, bufsiz, request_fmt, func, temp);
    free(temp);
}

void xmlrpc_request(char *buf, int bufsiz, const char *func, int narg, ...)
{
    va_list ap;
    va_start(ap, narg);
    xmlrpc_request_real(buf, bufsiz, func, narg, ap);
    va_end(ap);
}

void xmlrpc_request2(char *buf, int bufsiz, short int *buflen, const char *func, int narg, ...)
{
    va_list ap;
    va_start(ap, narg);
    xmlrpc_request_real(buf, bufsiz, func, narg, ap);
    *buflen = strlen(buf);
    va_end(ap);
}

