#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "xmlrpc_parse.h"

static int xmlrpc_grab(const char *inbuf, const char *tagnam, char *outbuf)
{
    char starttag[32];
    snprintf(starttag, sizeof(starttag), "<%s>", tagnam);
    char endtag[32];
    snprintf(endtag, sizeof(endtag), "</%s>", tagnam);
    char *start = strstr(inbuf, starttag);
    if(start == NULL) return 0;
    char *end;
    end = strstr(start, endtag);
    if(end == NULL) return 0;
    start += strlen(starttag);
    while(isspace(*start)) start++;
    while(isspace(*(end-1))) end--;
    strncpy(outbuf, start, end - start);
    outbuf[end - start] = 0;
    return strlen(starttag) + strlen(outbuf) + strlen(endtag);
}

static int xmlrpc_response_i4(const char *xmlstr, int *res)
{
    char *i4 = malloc(strlen(xmlstr) + 1);
    if(!xmlrpc_grab(xmlstr, "i4", i4)) return 0;
    *res = atoi(i4);
    free(i4);
    return 10;
}

static int xmlrpc_response_int(const char *xmlstr, int *res)
{
    char *xint = malloc(strlen(xmlstr) + 1);
    if(!xmlrpc_grab(xmlstr, "int", xint)) return 0;
    *res = atoi(xint);
    free(xint);
    return 10;
}

static int xmlrpc_response_double(const char *xmlstr, double *res)
{
    int len = strlen(xmlstr) + 1;
    char *xdouble = malloc(len);
    if(!xmlrpc_grab(xmlstr, "double", xdouble)) return 0;
    *res = atof(xdouble);
    free(xdouble);
    return 10;
}

static int xmlrpc_response_string(const char *xmlstr, char **res)
{
    char *string = malloc(strlen(xmlstr) + 1);
    if(!xmlrpc_grab(xmlstr, "string", string)) return 0;
    *res = string;
    return strlen(*res);
}

static int xmlrpc_response_value(const char *xmlstr, char **res)
{
    char *string = malloc(strlen(xmlstr) + 1);
    if(!xmlrpc_grab(xmlstr, "value", string)) return 0;
    *res = string;
    return strlen(*res);
}

static int xmlrpc_response_guess(const char *xmlstr, void *res)
{
    if(xmlrpc_response_i4(xmlstr, (int *)res)) return sizeof(int);
    if(xmlrpc_response_int(xmlstr, (int *)res)) return sizeof(int);
    if(xmlrpc_response_double(xmlstr, (double *)res)) return sizeof(double);
    if(xmlrpc_response_string(xmlstr, (char **)res)) return sizeof(char *);
    if(strstr(xmlstr, "<") == NULL)
    {
        char *string = malloc(strlen(xmlstr) + 1);
        strcpy(string, xmlstr);
        *((char **)res) = string;
        return sizeof(char *);
    }
    return 0;
}

int xmlrpc_response(const char *xmlstr, void *res)
{
    char *value = malloc(strlen(xmlstr) + 1);
    const char *p = xmlstr;
    int n = 0;
    for(;;) {
        char *p2 = strstr(p, "<member>");
        p = strstr(p, "<value>");
        if(p == NULL) break;
        if(strstr(p, "<array>") != NULL && strstr(p, "<array>") < strstr(p, "</value>")) {
            p = p + 1;
            p = strstr(p, "<value>");
        }
        if(strstr(p, "<struct>") != NULL && strstr(p, "<struct>") < strstr(p, "</value>")) {
            p = p + 1;
            p = strstr(p, "<value>");
        }
        if(p2 != NULL && p2 < p)  
        {
            {
                char *name = malloc(strlen(xmlstr) + 1);
                if(xmlrpc_grab(p2, "name", name))
                {
                    *((char **)res) = name;
                    char *res2 = res;
                    res2 += sizeof(char *);
                    res = res2;
                }
            }
        }
        int fndlen = xmlrpc_grab(p, "value", value);
        if(fndlen <= 0) break;
        {
            char *res2 = res;
            res2 += xmlrpc_response_guess(value, res);
            res = res2;
        }
        p = p + 1;
        n++;
    }
    free(value);
    return n;
}

int xmlrpc_response_status(const char *xmlstr)
{
    return strstr(xmlstr, "fault") == NULL;    
}

int xmlrpc_response_errmsg(const char *xmlstr, int *errcod, char **errmsg)
{
    if(!xmlrpc_response_i4(xmlstr, errcod)) xmlrpc_response_int(xmlstr, errcod);      
    return xmlrpc_response_string(xmlstr, errmsg);    
}

static struct struct_ctx
{
    int n;
    int nfld;
    struct xmlrpc_struct_i4_or_str *gas;
};

void *xmlrpc_struct_init(const char *xmlstr, int nfld)
{
    struct struct_ctx *ctx = malloc(sizeof(struct struct_ctx));
    ctx->gas = malloc(strlen(xmlstr) / 40 * sizeof(struct xmlrpc_struct_i4_or_str));
    ctx->nfld = nfld;
    ctx->n = xmlrpc_response(xmlstr, ctx->gas) / 2;
    return ctx;
}

int xmlrpc_struct_n(void *ctx)
{
    struct struct_ctx *ctx2 = ctx;
    return ctx2->n;
}

int xmlrpc_struct_unpack_i4(void *ctx, int ix, const char *name)
{
    struct struct_ctx *ctx2 = ctx;
    for(int i = 0; i < ctx2->nfld; i++)
    {
        if(strcmp(ctx2->gas[ix * ctx2->nfld + i].name, name) == 0)
        {
            return ctx2->gas[ix * ctx2->nfld + i].value.i4;
        }
    }
    return 0; 
}

char *xmlrpc_struct_unpack_string(void *ctx, int ix, const char *name)
{
    struct struct_ctx *ctx2 = ctx;
    for(int i = 0; i < ctx2->nfld; i++)
    {
        if(strcmp(ctx2->gas[ix * ctx2->nfld + i].name, name) == 0)
        {
            return ctx2->gas[ix * ctx2->nfld + i].value.str;
        }
    }
    return NULL;
}

void xmlrpc_struct_cleanup(void *ctx)
{
    struct struct_ctx *ctx2 = ctx;
    free(ctx2->gas);
    free(ctx2);
}

